from types import SimpleNamespace
from typing import Any, Callable, Dict, Optional, Sequence, Set, Tuple, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import inspect

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike

from darts.logging import get_logger, raise_if_not, raise_log
from darts.timeseries import TimeSeries
from darts.utils.ts_utils import get_series_seq_type, series2seq
from darts.utils.utils import generate_index

logger = get_logger(__name__)

TimeIndex = Union[
    pd.DatetimeIndex,
    pd.RangeIndex,
    Tuple[int, int],
    Tuple[pd.Timestamp, pd.Timestamp],
]


def _historical_forecasts_general_checks(model, series, kwargs):
    """
    Performs checks common to ForecastingModel and RegressionModel backtest() methods

    Parameters
    ----------
    model
        The forecasting model.
    series
        Either series when called from ForecastingModel, or target_series if called from RegressionModel
    signature_params
        A dictionary of the signature parameters of the calling method, to get the default values
        Typically would be signature(self.backtest).parameters
    kwargs
        Params specified by the caller of backtest(), they take precedence over the arguments' default values
    """
    # parse kwargs
    n = SimpleNamespace(**kwargs)

    # check forecast horizon
    raise_if_not(
        n.forecast_horizon > 0,
        "The provided forecasting horizon must be a positive integer.",
        logger,
    )

    # check stride
    raise_if_not(
        n.stride > 0,
        "The provided stride parameter must be a positive integer.",
        logger,
    )

    series = series2seq(series)

    if n.start is not None:
        # check start parameter in general (non series dependent)
        if not isinstance(n.start, (float, int, np.int64, pd.Timestamp)):
            raise_log(
                TypeError(
                    "`start` must be either `float`, `int`, `pd.Timestamp` or `None`."
                ),
                logger,
            )

        if n.start_format not in ["position", "value"]:
            raise_log(
                ValueError(
                    f"`start_format` must be on of ['position', 'value']. Received '{n.start_format}'."
                )
            )
        if n.start_format == "position":
            raise_if_not(
                isinstance(n.start, (int, np.int64)),
                f"Since `start_format='position'`, `start` must be an integer, received {type(n.start)}.",
                logger,
            )

        if isinstance(n.start, float):
            raise_if_not(
                0.0 <= n.start <= 1.0,
                "if `start` is a float, must be between 0.0 and 1.0.",
                logger,
            )

        # verbose error messages
        if not isinstance(n.start, pd.Timestamp):
            start_value_msg = f"`start` value `{n.start}` corresponding to timestamp"
        else:
            start_value_msg = "`start` time"
        for idx, series_ in enumerate(series):
            # check specifically for int and Timestamp as error by `get_timestamp_at_point` is too generic
            if isinstance(n.start, pd.Timestamp):
                if n.start > series_.end_time():
                    raise_log(
                        ValueError(
                            f"`start` time `{n.start}` is after the last timestamp `{series_.end_time()}` of the "
                            f"series at index: {idx}."
                        ),
                        logger,
                    )
                elif n.start < series_.start_time():
                    raise_log(
                        ValueError(
                            f"`start` time `{n.start}` is before the first timestamp `{series_.start_time()}` of the "
                            f"series at index: {idx}."
                        ),
                        logger,
                    )
            elif isinstance(n.start, (int, np.int64)):
                out_of_bound_error = False
                if n.start_format == "position":
                    if (n.start > 0 and n.start >= len(series_)) or (
                        n.start < 0 and np.abs(n.start) > len(series_)
                    ):
                        out_of_bound_error = True
                elif series_.has_datetime_index:
                    if n.start >= len(series_):
                        out_of_bound_error = True
                elif n.start < series_.time_index[0]:
                    raise_log(
                        ValueError(
                            f"`start` index `{n.start}` is smaller than the first index `{series_.time_index[0]}` "
                            f"for series at index: {idx}."
                        ),
                        logger,
                    )
                elif n.start > series_.time_index[-1]:
                    raise_log(
                        ValueError(
                            f"`start` index `{n.start}` is larger than the last index `{series_.time_index[-1]}` "
                            f"for series at index: {idx}."
                        ),
                        logger,
                    )

                if out_of_bound_error:
                    raise_log(
                        ValueError(
                            f"`start` index `{n.start}` is out of bounds for series of length {len(series_)} "
                            f"at index: {idx}."
                        ),
                        logger,
                    )

            if n.start_format == "value":
                start = series_.get_timestamp_at_point(n.start)
            else:
                start = series_.time_index[n.start]

            if n.retrain is not False and start == series_.start_time():
                raise_log(
                    ValueError(
                        f"{start_value_msg} `{start}` is the first timestamp of the series {idx}, resulting in an "
                        f"empty training set."
                    ),
                    logger,
                )

            # check that overlap_end and start together form a valid combination
            overlap_end = n.overlap_end
            if (
                not overlap_end
                and start + (series_.freq * (n.forecast_horizon - 1)) not in series_
            ):
                raise_log(
                    ValueError(
                        f"{start_value_msg} `{start}` is too late in the series {idx} to make any predictions with "
                        f"`overlap_end` set to `False`."
                    ),
                    logger,
                )

    # check direct likelihood parameter prediction before fitting a model
    if n.predict_likelihood_parameters:
        if not model.supports_likelihood_parameter_prediction:
            raise_log(
                ValueError(
                    f"Model `{model.__class__.__name__}` does not support `predict_likelihood_parameters=True`. "
                    f"Either the model does not support likelihoods, or no `likelihood` was used at model "
                    f"creation."
                )
            )
        if n.num_samples != 1:
            raise_log(
                ValueError(
                    f"`predict_likelihood_parameters=True` is only supported for `num_samples=1`, "
                    f"received {n.num_samples}."
                ),
                logger,
            )

        if (
            model.output_chunk_length is not None
            and n.forecast_horizon > model.output_chunk_length
        ):
            raise_log(
                ValueError(
                    "`predict_likelihood_parameters=True` is only supported for `forecast_horizon` smaller than or "
                    "equal to model's `output_chunk_length`."
                ),
                logger,
            )

    if (
        n.sample_weight is not None
        and not isinstance(n.sample_weight, str)
        and model.supports_sample_weight
    ):
        sample_weight = series2seq(n.sample_weight)
        for idx, (series_, sample_weight_) in enumerate(zip(series, sample_weight)):
            is_valid = (
                sample_weight_.freq == series_.freq
                and sample_weight_.start_time() <= series_.start_time()
                and len(sample_weight_) >= len(series_)
            )
            if not is_valid:
                raise_log(
                    ValueError(
                        f"`sample_weight` at series index {idx} must contain at least all times "
                        f"of the corresponding target `series`."
                    ),
                    logger=logger,
                )


def _historical_forecasts_sanitize_kwargs(
    model,
    fit_kwargs: Optional[Dict[str, Any]],
    predict_kwargs: Optional[Dict[str, Any]],
    retrain: bool,
    show_warnings: bool,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Convert kwargs to dictionary, check that their content is compatible with called methods."""
    hfc_args = set(inspect.signature(model.historical_forecasts).parameters)
    # replace `forecast_horizon` with `n`
    hfc_args = hfc_args - {"forecast_horizon"}
    hfc_args = hfc_args.union({"n"})

    if fit_kwargs is None:
        fit_kwargs = dict()
    elif retrain:
        fit_kwargs = _historical_forecasts_check_kwargs(
            hfc_args=hfc_args,
            name_kwargs="fit_kwargs",
            dict_kwargs=fit_kwargs,
        )
    elif show_warnings:
        logger.warning(
            "`fit_kwargs` was provided with `retrain=False`, the argument will be ignored."
        )

    if predict_kwargs is None:
        predict_kwargs = dict()
    else:
        predict_kwargs = _historical_forecasts_check_kwargs(
            hfc_args=hfc_args,
            name_kwargs="predict_kwargs",
            dict_kwargs=predict_kwargs,
        )
    return fit_kwargs, predict_kwargs


def _historical_forecasts_check_kwargs(
    hfc_args: Set[str],
    name_kwargs: str,
    dict_kwargs: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Return the kwargs dict without the arguments unsupported by the model method.

    Raise a warning if some argument are not supported and an exception if some arguments interfere with
    historical_forecasts logic.
    """
    invalid_args = set(dict_kwargs).intersection(hfc_args)
    if len(invalid_args) > 0:
        raise_log(
            ValueError(
                f"The following parameters cannot be passed in `{name_kwargs}`: {invalid_args}. "
                f"Make sure to pass them explicitly to the function/method."
            ),
            logger,
        )
    return dict_kwargs


def _historical_forecasts_start_warnings(
    idx: int,
    start: Union[pd.Timestamp, int],
    start_time_: Union[int, pd.Timestamp],
    historical_forecasts_time_index: TimeIndex,
):
    """Warnings when start value provided by user is not within the forecastable indexes boundaries"""
    if not isinstance(start, pd.Timestamp):
        start_value_msg = f"value `{start}` corresponding to timestamp `{start_time_}`"
    else:
        start_value_msg = f"time `{start_time_}`"

    if start_time_ < historical_forecasts_time_index[0]:
        logger.warning(
            f"`start` {start_value_msg} is before the first predictable/trainable historical "
            f"forecasting point for series at index: {idx}. Ignoring `start` for this series and "
            f"beginning at first trainable/predictable time: {historical_forecasts_time_index[0]}. "
            f"To hide these warnings, set `show_warnings=False`."
        )
    else:
        logger.warning(
            f"`start` {start_value_msg} is after the last trainable/predictable historical "
            f"forecasting point for series at index: {idx}. This would results in empty historical "
            f"forecasts. Ignoring `start` for this series and beginning at first trainable/"
            f"predictable time: {historical_forecasts_time_index[0]}. Non-empty forecasts can be "
            f"generated by setting `start` value to times between (including): "
            f"{(historical_forecasts_time_index[0], historical_forecasts_time_index[-1])}. "
            f"To hide these warnings, set `show_warnings=False`."
        )


def _get_historical_forecastable_time_index(
    model,
    series: TimeSeries,
    forecast_horizon: int,
    overlap_end: bool,
    past_covariates: Optional[TimeSeries] = None,
    future_covariates: Optional[TimeSeries] = None,
    is_training: Optional[bool] = False,
    reduce_to_bounds: bool = False,
) -> Union[
    pd.DatetimeIndex,
    pd.RangeIndex,
    Tuple[int, int],
    Tuple[pd.Timestamp, pd.Timestamp],
    None,
]:
    """
    Private function that returns the largest time_index representing the subset of each timestamps
    for which historical forecasts can be made, given the model's properties, the training series
    and the covariates.
        - If ``None`` is returned, there is no point where a forecast can be made.

        - If ``is_training=False``, it returns the time_index subset of predictable timestamps.

        - If ``is_training=True``, it returns the time_index subset of trainable timestamps. A trainable
        timestamp is a timestamp that has a training sample of length at least ``self.training_sample_length``
            preceding it.

    Additionally, it accounts for auto-regression (forecast_horizon > model.output_chunk_length , and overlap_end.
        - In case of auto-regression, we have to step back the last possible predictable/trainable time step if the
          covariates are too short
        - In case overlap_end=False, the latest possible predictable/trainable time step is shifted back if a
          prediction starting from that point would go beyond the end of `series`.


    Parameters
    ----------
    series
        A target series.
    forecast_horizon
        The forecast horizon for the predictions.
    overlap_end
        Whether the returned forecasts can go beyond the series' end or not.
    past_covariates
        Optionally, a past covariates.
    future_covariates
        Optionally, a future covariates.
    is_training
        Whether the returned time_index should be taking into account the training.
    reduce_to_bounds
        Whether to only return the minimum and maximum historical forecastable index

    Returns
    -------
    Union[pd.DatetimeIndex, pd.RangeIndex, Tuple[int, int], Tuple[pd.Timestamp, pd.Timestamp], None]
        The longest time_index that can be used for historical forecasting, either as a range or a tuple.

    Examples
    --------
    >>> model = LinearRegressionModel(lags=3, output_chunk_length=2)
    >>> model.fit(train_series)
    >>> series = TimeSeries.from_times_and_values(pd.date_range('2000-01-01', '2000-01-10'), np.arange(10))
    >>> model._get_historical_forecastable_time_index(series=series, is_training=False, forecast_horizon=1)
    DatetimeIndex(
            ['2000-01-04', '2000-01-05', '2000-01-06', '2000-01-07', '2000-01-08', '2000-01-09', '2000-01-10'],
            dtype='datetime64[ns]', freq='D'
    )
    >>> model._get_historical_forecastable_time_index(series=series, is_training=True)
    DatetimeIndex(['2000-01-06', '2000-01-08', '2000-01-09', '2000-01-10'], dtype='datetime64[ns]', freq='D')
    >>> model = NBEATSModel(input_chunk_length=3, output_chunk_length=3)
    >>> model.fit(train_series, train_past_covariates)
    >>> series = TimeSeries.from_times_and_values(pd.date_range('2000-10-01', '2000-10-09'), np.arange(8))
    >>> past_covariates = TimeSeries.from_times_and_values(
    >>>     pd.date_range('2000-10-03', '2000-10-20'),
    >>>     np.arange(18)
    >>> )
    >>> model._get_historical_forecastable_time_index(
    >>>     series=series,
    >>>     past_covariates=past_covariates,
    >>>     is_training=False,
    >>>     forecast_horizon=1,
    >>> )
    DatetimeIndex(['2000-10-06', '2000-10-07', '2000-10-08', '2000-10-09'], dtype='datetime64[ns]', freq='D')
    >>>  # Only one point is trainable; it corresponds to the first point after we reach a common subset of
    >>> # timestamps of training_sample_length length.
    >>> model._get_historical_forecastable_time_index(
    >>>     series=series,
    >>>     past_covariates=past_covariates,
    >>>     is_training=True,
    >>> )
    DatetimeIndex(['2000-10-09'], dtype='datetime64[ns]', freq='D')
    """

    (
        min_target_lag,
        max_target_lag,
        min_past_cov_lag,
        max_past_cov_lag,
        min_future_cov_lag,
        max_future_cov_lag,
        output_chunk_shift,
        max_target_lag_train,
    ) = model.extreme_lags

    # max_target_lag < 0 are local models which can predict for n (horizon) -> infinity (no auto-regression)
    is_autoregression = (
        max_target_lag >= 0
        and forecast_horizon > max_target_lag - output_chunk_shift + 1
    )

    if min_target_lag is None:
        min_target_lag = 0

    if is_training and max_target_lag_train is not None:
        # the output lag/window can be different for train and predict modes
        output_lag = max_target_lag_train
    else:
        output_lag = max_target_lag

    # longest possible time index for target
    if is_training:
        start = (
            series.start_time()
            + (output_lag - output_chunk_shift - min_target_lag + 1) * series.freq
        )
    else:
        start = series.start_time() - min_target_lag * series.freq
    end = series.end_time() + 1 * series.freq

    intersect_ = (start, end)

    # longest possible time index for past covariates
    if (min_past_cov_lag is not None) and (past_covariates is not None):
        if is_training:
            start_pc = (
                past_covariates.start_time()
                + (output_lag - output_chunk_shift - min_past_cov_lag + 1)
                * past_covariates.freq
            )
        else:
            start_pc = (
                past_covariates.start_time() - min_past_cov_lag * past_covariates.freq
            )

        shift_pc_end = max_past_cov_lag
        if is_autoregression:
            # we step back in case of auto-regression
            shift_pc_end += forecast_horizon - (max_target_lag - output_chunk_shift + 1)
        end_pc = past_covariates.end_time() - shift_pc_end * past_covariates.freq

        intersect_ = (
            max([intersect_[0], start_pc]),
            min([intersect_[1], end_pc]),
        )

    # longest possible time index for future covariates
    if (min_future_cov_lag is not None) and (future_covariates is not None):
        if is_training:
            start_fc = (
                future_covariates.start_time()
                + (output_lag - output_chunk_shift - min_future_cov_lag + 1)
                * future_covariates.freq
            )
        else:
            start_fc = (
                future_covariates.start_time()
                - min_future_cov_lag * future_covariates.freq
            )

        shift_fc_end = max_future_cov_lag
        if is_autoregression:
            # we step back in case of auto-regression
            shift_fc_end += forecast_horizon - (max_target_lag - output_chunk_shift + 1)
        end_fc = future_covariates.end_time() - shift_fc_end * future_covariates.freq

        intersect_ = (
            max([intersect_[0], start_fc]),
            min([intersect_[1], end_fc]),
        )

    # overlap_end = False -> predictions must not go beyond end of target series
    if (
        not overlap_end
        and intersect_[1] + (forecast_horizon + output_chunk_shift - 1) * series.freq
        > series.end_time()
    ):
        intersect_ = (
            intersect_[0],
            end - (forecast_horizon + output_chunk_shift) * series.freq,
        )

    # end comes before the start
    if intersect_[1] < intersect_[0]:
        return None

    # if RegressionModel is not multi_models, it looks further in the past
    is_multi_models = getattr(model, "multi_models", None)
    if is_multi_models is not None and not is_multi_models:
        intersect_ = (
            intersect_[0] + (model.output_chunk_length - 1) * series.freq,
            intersect_[1],
        )

    # generate an index
    if not reduce_to_bounds:
        intersect_ = generate_index(
            start=intersect_[0], end=intersect_[1], freq=series.freq
        )

    return intersect_ if len(intersect_) > 0 else None


def _adjust_historical_forecasts_time_index(
    series: TimeSeries,
    series_idx: int,
    historical_forecasts_time_index: TimeIndex,
    start: Optional[Union[pd.Timestamp, float, int]],
    start_format: Literal["position", "value"],
    show_warnings: bool,
) -> TimeIndex:
    """
    Shrink the beginning and end of the historical forecasts time index based on the values of `start`,
    `forecast_horizon` and `overlap_end`.
    """
    # when applicable, shift the start of the forecastable index based on `start`
    if start is not None:
        if start_format == "value":
            start_time_ = series.get_timestamp_at_point(start)
        else:
            start_time_ = series.time_index[start]
        # ignore user-defined `start`
        if (
            not historical_forecasts_time_index[0]
            <= start_time_
            <= historical_forecasts_time_index[-1]
        ):
            if show_warnings:
                _historical_forecasts_start_warnings(
                    idx=series_idx,
                    start=start,
                    start_time_=start_time_,
                    historical_forecasts_time_index=historical_forecasts_time_index,
                )
        else:
            historical_forecasts_time_index = (
                max(historical_forecasts_time_index[0], start_time_),
                historical_forecasts_time_index[1],
            )

    return historical_forecasts_time_index


def _get_historical_forecast_predict_index(
    model,
    series: TimeSeries,
    series_idx: int,
    past_covariates: Optional[TimeSeries],
    future_covariates: Optional[TimeSeries],
    forecast_horizon: int,
    overlap_end: bool,
) -> TimeIndex:
    """Obtain the boundaries of the predictable time indices, raise an exception if None"""
    historical_forecasts_time_index = _get_historical_forecastable_time_index(
        model=model,
        series=series,
        forecast_horizon=forecast_horizon,
        overlap_end=overlap_end,
        past_covariates=past_covariates,
        future_covariates=future_covariates,
        is_training=False,
        reduce_to_bounds=True,
    )

    if historical_forecasts_time_index is None:
        raise_log(
            ValueError(
                "Cannot build a single input for prediction with the provided model, "
                f"`series` and `*_covariates` at series index: {series_idx}. The minimum "
                "prediction input time index requirements were not met. "
                "Please check the time index of `series` and `*_covariates`."
            )
        )

    return historical_forecasts_time_index


def _get_historical_forecast_train_index(
    model,
    series: TimeSeries,
    series_idx: int,
    past_covariates: Optional[TimeSeries],
    future_covariates: Optional[TimeSeries],
    forecast_horizon: int,
    overlap_end: bool,
) -> TimeIndex:
    """
    Obtain the boundaries of the time indices usable for training, raise an exception if training is required and
    no indices are available.
    """
    historical_forecasts_time_index = _get_historical_forecastable_time_index(
        model=model,
        series=series,
        forecast_horizon=forecast_horizon,
        overlap_end=overlap_end,
        past_covariates=past_covariates,
        future_covariates=future_covariates,
        is_training=True,
        reduce_to_bounds=True,
    )

    if not model._fit_called and historical_forecasts_time_index is None:
        raise_log(
            ValueError(
                "Cannot build a single input for training with the provided untrained model, "
                f"`series` and `*_covariates` at series index: {series_idx}. The minimum "
                "training input time index requirements were not met. "
                "Please check the time index of `series` and `*_covariates`."
            ),
            logger,
        )

    return historical_forecasts_time_index


def _reconciliate_historical_time_indices(
    model,
    historical_forecasts_time_index_predict: TimeIndex,
    historical_forecasts_time_index_train: TimeIndex,
    series: TimeSeries,
    series_idx: int,
    retrain: Union[bool, int, Callable[..., bool]],
    train_length: Optional[int],
    show_warnings: bool,
) -> Tuple[TimeIndex, Optional[int]]:
    """Depending on the value of retrain, select which time indices will be used during the historical forecasts."""
    train_length_ = None
    if isinstance(retrain, Callable):
        # retain the longer time index, anything can happen
        if (
            historical_forecasts_time_index_train is not None
            and historical_forecasts_time_index_train[0]
            < historical_forecasts_time_index_predict[0]
        ):
            historical_forecasts_time_index = historical_forecasts_time_index_train
        else:
            historical_forecasts_time_index = historical_forecasts_time_index_predict
    elif retrain:
        historical_forecasts_time_index = historical_forecasts_time_index_train
    else:
        historical_forecasts_time_index = historical_forecasts_time_index_predict

    # compute the maximum forecasts time index assuming that `start=None`
    if retrain or (not model._fit_called):
        if train_length and train_length <= len(series):
            train_length_ = train_length
            # we have to start later for larger `train_length`
            step_ahead = max(train_length - model._training_sample_time_index_length, 0)
            if step_ahead:
                historical_forecasts_time_index = (
                    historical_forecasts_time_index[0] + step_ahead * series.freq,
                    historical_forecasts_time_index[-1],
                )

        # if not we start training right away; some models (sklearn) require more than 1
        # training samples, so we start after the first trainable point.
        else:
            if train_length and train_length > len(series) and show_warnings:
                logger.warning(
                    f"`train_length` is larger than the length of series at index: {series_idx}. "
                    f"Ignoring `train_length` and using default behavior where all available time steps up "
                    f"until the end of the expanding training set. "
                    f"To hide these warnings, set `show_warnings=False`."
                )

            train_length_ = None
            if model.min_train_samples > 1:
                historical_forecasts_time_index = (
                    historical_forecasts_time_index[0]
                    + (model.min_train_samples - 1) * series.freq,
                    historical_forecasts_time_index[1],
                )

    return historical_forecasts_time_index, train_length_


def _get_historical_forecast_boundaries(
    model,
    series: TimeSeries,
    series_idx: int,
    past_covariates: Optional[TimeSeries],
    future_covariates: Optional[TimeSeries],
    start: Optional[Union[pd.Timestamp, float, int]],
    start_format: Literal["position", "value"],
    forecast_horizon: int,
    overlap_end: bool,
    freq: pd.DateOffset,
    show_warnings: bool = True,
) -> Tuple[Any, ...]:
    """
    Based on the boundaries of the forecastable time index, generates the boundaries of each covariates using the lags.

    For TimeSeries with a RangeIndex, the boundaries are converted to positional indexes to slice the array
    appropriately when start > 0.

    When applicable, move the start boundaries to the value provided by the user.
    """
    # obtain forecastable indexes boundaries, as values from the time index
    historical_forecasts_time_index = _get_historical_forecast_predict_index(
        model,
        series,
        series_idx,
        past_covariates,
        future_covariates,
        forecast_horizon,
        overlap_end,
    )

    # adjust boundaries based on start, forecast_horizon and overlap_end
    historical_forecasts_time_index = _adjust_historical_forecasts_time_index(
        series=series,
        series_idx=series_idx,
        historical_forecasts_time_index=historical_forecasts_time_index,
        start=start,
        start_format=start_format,
        show_warnings=show_warnings,
    )

    # re-adjust the slicing indexes to account for the lags
    # `max_target_lag_train` is redundant, since optimized hist fc is running in predict mode only
    (
        min_target_lag,
        _,
        min_past_cov_lag,
        max_past_cov_lag,
        min_future_cov_lag,
        max_future_cov_lag,
        output_chunk_shift,
        max_target_lag_train,
    ) = model.extreme_lags

    # target lags are <= 0
    hist_fct_tgt_start, hist_fct_tgt_end = historical_forecasts_time_index
    if min_target_lag is not None:
        hist_fct_tgt_start += min_target_lag * freq
    hist_fct_tgt_end -= 1 * freq
    # past lags are <= 0
    hist_fct_pc_start, hist_fct_pc_end = historical_forecasts_time_index
    if min_past_cov_lag is not None:
        hist_fct_pc_start += min_past_cov_lag * freq
    if max_past_cov_lag is not None:
        hist_fct_pc_end += max_past_cov_lag * freq
    # future lags can be anything
    hist_fct_fc_start, hist_fct_fc_end = historical_forecasts_time_index
    if min_future_cov_lag is not None:
        hist_fct_fc_start += min_future_cov_lag * freq
    if max_future_cov_lag is not None:
        hist_fct_fc_end += max_future_cov_lag * freq

    # convert actual integer index values (points) to positional index, make end bound inclusive
    if series.has_range_index:
        hist_fct_tgt_start = series.get_index_at_point(hist_fct_tgt_start)
        hist_fct_tgt_end = series.get_index_at_point(hist_fct_tgt_end) + 1
        if past_covariates is not None:
            hist_fct_pc_start = past_covariates.get_index_at_point(hist_fct_pc_start)
            hist_fct_pc_end = past_covariates.get_index_at_point(hist_fct_pc_end) + 1
        else:
            hist_fct_pc_start, hist_fct_pc_end = None, None
        if future_covariates is not None:
            hist_fct_fc_start = future_covariates.get_index_at_point(hist_fct_fc_start)
            hist_fct_fc_end = future_covariates.get_index_at_point(hist_fct_fc_end) + 1
        else:
            hist_fct_fc_start, hist_fct_fc_end = None, None

    return (
        historical_forecasts_time_index[0],
        historical_forecasts_time_index[1],
        hist_fct_tgt_start,
        hist_fct_tgt_end,
        hist_fct_pc_start,
        hist_fct_pc_end,
        hist_fct_fc_start,
        hist_fct_fc_end,
    )


def _check_optimizable_historical_forecasts_global_models(
    model,
    forecast_horizon: int,
    retrain: Union[bool, int, Callable[..., bool]],
    show_warnings: bool,
    allow_autoregression: bool,
) -> bool:
    """
    Historical forecast can be optimized only if `retrain=False`. If `allow_autoregression=False`, historical forecasts
    can be optimized only if `forecast_horizon <= model.output_chunk_length` (no auto-regression required).
    """

    retrain_off = (retrain is False) or (retrain == 0)
    is_autoregressive = forecast_horizon > model.output_chunk_length
    if retrain_off and (
        not is_autoregressive or (is_autoregressive and allow_autoregression)
    ):
        return True

    if show_warnings:
        if not retrain_off:
            logger.warning(
                "`enable_optimization=True` is ignored because `retrain` is not `False` or `0`. "
                "To hide this warning, set `show_warnings=False` or `enable_optimization=False`."
            )
        if is_autoregressive:
            logger.warning(
                "`enable_optimization=True` is ignored because `forecast_horizon > model.output_chunk_length`. "
                "To hide this warning, set `show_warnings=False` or `enable_optimization=False`."
            )

    return False


def _process_historical_forecast_input(
    model,
    series: Union[TimeSeries, Sequence[TimeSeries]],
    past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    forecast_horizon: int = 1,
    allow_autoregression: bool = False,
) -> Union[
    Sequence[TimeSeries],
    Optional[Sequence[TimeSeries]],
    Optional[Sequence[TimeSeries]],
    int,
]:
    if not model._fit_called:
        raise_log(
            ValueError("Model has not been fit yet."),
            logger,
        )

    if not allow_autoregression and forecast_horizon > model.output_chunk_length:
        raise_log(
            ValueError(
                "`forecast_horizon > model.output_chunk_length` requires auto-regression which is not "
                "supported in this optimized routine."
            ),
            logger,
        )
    series_seq_type = get_series_seq_type(series)
    series = series2seq(series)
    past_covariates = series2seq(past_covariates)
    future_covariates = series2seq(future_covariates)

    # manage covariates, usually handled by RegressionModel.predict()
    if past_covariates is None and model.past_covariate_series is not None:
        past_covariates = [model.past_covariate_series] * len(series)
    if future_covariates is None and model.future_covariate_series is not None:
        future_covariates = [model.future_covariate_series] * len(series)

    if model.uses_static_covariates:
        model._verify_static_covariates(series[0].static_covariates)

    if model.encoders.encoding_available:
        past_covariates, future_covariates = model.generate_fit_predict_encodings(
            n=forecast_horizon,
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )
    return series, past_covariates, future_covariates, series_seq_type


def _process_predict_start_points_bounds(
    series: Sequence[TimeSeries], bounds: ArrayLike, stride: int
) -> Tuple[np.ndarray, np.ndarray]:
    """Processes the historical forecastable time index bounds (earliest, and latest possible prediction
    start points).

    Parameters
    ----------
    bounds
        An array of shape (n series, 2), with the left and right prediction start point bounds per series.
    stride
        The number of time steps between two consecutive predictions.

    Returns
    -------
    (np.ndarray, np.ndarray)
        The adjusted bounds: the right bounds are adjusted to be a multiple of 'stride' ahead of the left bounds.
        The number of resulting predicted series per input series respecting stride and bounds.
    """
    bounds = bounds if isinstance(bounds, np.ndarray) else np.array(bounds)
    if not bounds.shape == (len(series), 2):
        raise_log(
            ValueError(
                "`bounds` must be an array like with shape `(n target series, 2)`, "
                "with the start and end bounds of each series"
            ),
            logger=logger,
        )
    # we might have some steps that are too long considering stride
    steps_too_long = (bounds[:, 1] - bounds[:, 0]) % stride
    bounds[:, 1] -= steps_too_long
    cum_lengths = np.cumsum(np.diff(bounds) // stride + 1)
    return bounds, cum_lengths



"""
Forecasting Model Explainer Base Class

A `_ForecastingModelExplainer` takes a fitted forecasting model as input and generates explanations for it.
"""

from abc import ABC, abstractmethod
from typing import Optional, Sequence, Tuple, Union

from darts import TimeSeries
from darts.explainability.explainability_result import _ExplainabilityResult
from darts.explainability.utils import process_horizons_and_targets, process_input
from darts.logging import get_logger, raise_log
from darts.models.forecasting.forecasting_model import ForecastingModel

logger = get_logger(__name__)

MIN_BACKGROUND_SAMPLE = 10


class _ForecastingModelExplainer(ABC):
    @abstractmethod
    def __init__(
        self,
        model: ForecastingModel,
        background_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        background_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        background_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        requires_background: bool = False,
        requires_covariates_encoding: bool = False,
        check_component_names: bool = False,
        test_stationarity: bool = False,
    ):
        """
        The base class for forecasting model explainers. It defines the *minimal* behavior that all
        forecasting model explainers support.

        Naming:

        - A background series is a `TimeSeries` with which to 'train' the `Explainer` model.
        - A foreground series is the `TimeSeries` to explain using the fitted `Explainer` model.

        Parameters
        ----------
        model
            A `ForecastingModel` to be explained. It must be fitted first.
        background_series
            A series or list of series to *train* the `_ForecastingModelExplainer` along with any foreground series.
            Consider using a reduced well-chosen background to reduce computation time.
            Optional if `model` was fit on a single target series. By default, it is the `series` used at fitting time.
            Mandatory if `model` was fit on multiple (sequence of) target series.
        background_past_covariates
            A past covariates series or list of series that the model needs once fitted.
        background_future_covariates
            A future covariates series or list of series that the model needs once fitted.
        requires_background
            Whether the explainer requires background series as an input. If `True`, raises an error if no background
            series were provided and `model` was fit using multiple series.
        requires_covariates_encoding
            Whether to apply the model's encoders to the input covariates. This should only be `True` if the
            Explainer will not call model methods `fit()` or `predict()` directly.
        check_component_names
            Whether to enforce that, in the case of multiple time series, all series of the same type (target and/or
            *_covariates) must have the same component names.
        test_stationarity
            Whether to raise a warning if not all `background_series` are stationary.
        """
        if not model._fit_called:
            raise_log(
                ValueError(
                    f"The model must be fitted before instantiating a {self.__class__.__name__}."
                ),
                logger,
            )
        self.model = model
        # default forecasting horizon
        self.n: Optional[int] = getattr(self.model, "output_chunk_length", None)

        # check background input validity and process it
        (
            self.background_series,
            self.background_past_covariates,
            self.background_future_covariates,
            self.target_components,
            self.static_covariates_components,
            self.past_covariates_components,
            self.future_covariates_components,
        ) = process_input(
            model=model,
            input_type="background",
            series=background_series,
            past_covariates=background_past_covariates,
            future_covariates=background_future_covariates,
            fallback_series=model.training_series,
            fallback_past_covariates=model.past_covariate_series,
            fallback_future_covariates=model.future_covariate_series,
            check_component_names=check_component_names,
            requires_input=requires_background,
            requires_covariates_encoding=requires_covariates_encoding,
            test_stationarity=test_stationarity,
        )
        self.requires_foreground = self.background_series is None
        self.requires_covariates_encoding = requires_covariates_encoding
        self.check_component_names = check_component_names
        self.test_stationarity = test_stationarity

    @abstractmethod
    def explain(
        self,
        foreground_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        foreground_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        foreground_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        horizons: Optional[Sequence[int]] = None,
        target_components: Optional[Sequence[str]] = None,
    ) -> _ExplainabilityResult:
        """
        Explains a foreground time series, and returns a :class:`_ExplainabilityResult
        <darts.explainability.explainability_result._ExplainabilityResult>` that can be used for downstream tasks.

        Parameters
        ----------
        foreground_series
            Optionally, one or a sequence of target `TimeSeries` to be explained. Can be multivariate.
            If not provided, the background `TimeSeries` will be explained instead.
        foreground_past_covariates
            Optionally, one or a sequence of past covariates `TimeSeries` if required by the forecasting model.
        foreground_future_covariates
            Optionally, one or a sequence of future covariates `TimeSeries` if required by the forecasting model.
        horizons
            Optionally, an integer or sequence of integers representing the future time steps to be explained.
            `1` corresponds to the first timestamp being forecasted.
            All values must be `<=output_chunk_length` of the explained forecasting model.
        target_components
            Optionally, a string or sequence of strings with the target components to explain.

        Returns
        -------
        _ExplainabilityResult
            The explainability result.
        """
        pass

    def _process_foreground(
        self,
        foreground_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        foreground_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        foreground_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
    ):
        return process_input(
            model=self.model,
            input_type="foreground",
            series=foreground_series,
            past_covariates=foreground_past_covariates,
            future_covariates=foreground_future_covariates,
            fallback_series=self.background_series,
            fallback_past_covariates=self.background_past_covariates,
            fallback_future_covariates=self.background_future_covariates,
            check_component_names=self.check_component_names,
            requires_input=self.requires_foreground,
            requires_covariates_encoding=self.requires_covariates_encoding,
            test_stationarity=self.test_stationarity,
        )

    def _process_horizons_and_targets(
        self,
        horizons: Optional[Union[int, Sequence[int]]],
        target_components: Optional[Union[str, Sequence[str]]],
    ) -> Tuple[Sequence[int], Sequence[str]]:
        return process_horizons_and_targets(
            horizons=horizons,
            fallback_horizon=self.n,
            target_components=target_components,
            fallback_target_components=self.target_components,
            check_component_names=self.check_component_names,
        )



from itertools import product
from typing import Dict, Sequence, Tuple

import numpy as np
import pandas as pd
import pytest
from pyod.models.knn import KNN

from darts import TimeSeries
from darts.ad import (
    AndAggregator,  # noqa: F401
    CauchyNLLScorer,
    EnsembleSklearnAggregator,  # noqa: F401
    ExponentialNLLScorer,
    FilteringAnomalyModel,
    ForecastingAnomalyModel,
    GammaNLLScorer,
    GaussianNLLScorer,
    KMeansScorer,
    LaplaceNLLScorer,
    OrAggregator,  # noqa: F401
    PoissonNLLScorer,
    PyODScorer,
    QuantileDetector,  # noqa: F401
    ThresholdDetector,  # noqa: F401
    WassersteinScorer,
)
from darts.ad import DifferenceScorer as Difference
from darts.ad import NormScorer as Norm
from darts.ad.utils import eval_metric_from_scores, show_anomalies_from_scores
from darts.models import MovingAverageFilter, NaiveSeasonal, RegressionModel

filtering_am = [
    (
        FilteringAnomalyModel,
        {"model": MovingAverageFilter(window=10), "scorer": Norm()},
    ),
    (
        FilteringAnomalyModel,
        {"model": MovingAverageFilter(window=10), "scorer": [Norm(), KMeansScorer()]},
    ),
    (
        FilteringAnomalyModel,
        {"model": MovingAverageFilter(window=10), "scorer": KMeansScorer()},
    ),
]

forecasting_am = [
    (ForecastingAnomalyModel, {"model": RegressionModel(lags=10), "scorer": Norm()}),
    (
        ForecastingAnomalyModel,
        {"model": RegressionModel(lags=10), "scorer": [Norm(), KMeansScorer()]},
    ),
    (
        ForecastingAnomalyModel,
        {"model": RegressionModel(lags=10), "scorer": KMeansScorer()},
    ),
]


class TestAnomalyDetectionModel:
    np.random.seed(42)

    # univariate series
    np_train = np.random.normal(loc=10, scale=0.5, size=100)
    train = TimeSeries.from_values(np_train)

    np_covariates = np.random.choice(a=[0, 1], size=100, p=[0.5, 0.5])
    covariates = TimeSeries.from_times_and_values(train._time_index, np_covariates)

    np_test = np.random.normal(loc=10, scale=1, size=100)
    test = TimeSeries.from_times_and_values(train._time_index, np_test)

    np_anomalies = np.random.choice(a=[0, 1], size=100, p=[0.9, 0.1])
    anomalies = TimeSeries.from_times_and_values(train._time_index, np_anomalies)

    np_only_1_anomalies = np.random.choice(a=[0, 1], size=100, p=[0, 1])
    only_1_anomalies = TimeSeries.from_times_and_values(
        train._time_index, np_only_1_anomalies
    )

    np_only_0_anomalies = np.random.choice(a=[0, 1], size=100, p=[1, 0])
    only_0_anomalies = TimeSeries.from_times_and_values(
        train._time_index, np_only_0_anomalies
    )

    modified_train = MovingAverageFilter(window=10).filter(train)
    modified_test = MovingAverageFilter(window=10).filter(test)

    np_probabilistic = np.random.normal(loc=10, scale=1, size=[100, 1, 20])
    probabilistic = TimeSeries.from_times_and_values(
        train._time_index, np_probabilistic
    )

    # multivariate series
    np_mts_train = np.random.normal(loc=[10, 5], scale=[0.5, 1], size=[100, 2])
    mts_train = TimeSeries.from_values(np_mts_train)

    np_mts_test = np.random.normal(loc=[10, 5], scale=[1, 1.5], size=[100, 2])
    mts_test = TimeSeries.from_times_and_values(mts_train._time_index, np_mts_test)

    np_mts_anomalies = np.random.choice(a=[0, 1], size=[100, 2], p=[0.9, 0.1])
    mts_anomalies = TimeSeries.from_times_and_values(
        mts_train._time_index, np_mts_anomalies
    )

    @pytest.mark.parametrize(
        "scorer,anomaly_model_config",
        product(
            [
                Norm(),
                Difference(),
                GaussianNLLScorer(),
                ExponentialNLLScorer(),
                PoissonNLLScorer(),
                LaplaceNLLScorer(),
                CauchyNLLScorer(),
                GammaNLLScorer(),
            ],
            [
                (ForecastingAnomalyModel, {"model": RegressionModel(lags=10)}),
                (FilteringAnomalyModel, {"model": MovingAverageFilter(window=20)}),
            ],
        ),
    )
    def test_non_fittable_scorer(self, scorer, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(scorer=scorer, **am_kwargs)
        assert not anomaly_model.scorers_are_trainable

    @pytest.mark.parametrize(
        "scorer,anomaly_model_config",
        product(
            [
                PyODScorer(model=KNN()),
                KMeansScorer(),
                WassersteinScorer(window_agg=False),
            ],
            [
                (ForecastingAnomalyModel, {"model": RegressionModel(lags=10)}),
                (FilteringAnomalyModel, {"model": MovingAverageFilter(window=20)}),
            ],
        ),
    )
    def test_fittable_scorer(self, scorer, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(scorer=scorer, **am_kwargs)
        assert anomaly_model.scorers_are_trainable

    def test_no_local_model(self):
        with pytest.raises(ValueError) as err:
            _ = ForecastingAnomalyModel(model=NaiveSeasonal(), scorer=KMeansScorer())
        assert str(err.value) == "`model` must be a Darts `GlobalForecastingModel`."

    @pytest.mark.parametrize(
        "anomaly_model,fit_model",
        [
            (
                ForecastingAnomalyModel(model=RegressionModel(lags=10), scorer=Norm()),
                True,
            ),
            (
                FilteringAnomalyModel(
                    model=MovingAverageFilter(window=20), scorer=Norm()
                ),
                False,
            ),
        ],
    )
    def test_score(self, anomaly_model, fit_model):
        if fit_model:
            anomaly_model.fit(self.train, allow_model_training=True)

        # if return_model_prediction set to true, output must be tuple
        assert isinstance(
            anomaly_model.score(self.test, return_model_prediction=True), Tuple
        )

        # if return_model_prediction set to false output must be
        # Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]
        assert not isinstance(
            anomaly_model.score(self.test, return_model_prediction=False), Tuple
        )

    @pytest.mark.parametrize("anomaly_model_config", filtering_am)
    def test_FitFilteringAnomalyModelInput(self, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(**am_kwargs)
        # `allow_model_training=True` has no effect if filter model has no `fit()` method
        anomaly_model.fit(self.train, allow_model_training=True)

        # input 'series' must be a series or Sequence of series
        with pytest.raises(ValueError):
            anomaly_model.fit([self.train, "str"], allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit([[self.train, self.train]], allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit("str", allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit([1, 2, 3], allow_model_training=True)

    @pytest.mark.parametrize("anomaly_model_config", forecasting_am)
    def test_FitForecastingAnomalyModelInput(self, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(**am_kwargs)

        # input 'series' must be a series or Sequence of series
        with pytest.raises(ValueError):
            anomaly_model.fit([self.train, "str"], allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit([[self.train, self.train]], allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit("str", allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit([1, 2, 3], allow_model_training=True)

        # 'allow_model_training' must be set to True if forecasting model is not fitted
        if anomaly_model.scorers_are_trainable:
            with pytest.raises(ValueError):
                anomaly_model.fit(self.train, allow_model_training=False)
                anomaly_model.score(self.train)

        with pytest.raises(ValueError):
            # number of 'past_covariates' must be the same as the number of Timeseries in 'series'
            anomaly_model.fit(
                series=[self.train, self.train],
                past_covariates=self.covariates,
                allow_model_training=True,
            )

        with pytest.raises(ValueError):
            # number of 'past_covariates' must be the same as the number of Timeseries in 'series'
            anomaly_model.fit(
                series=self.train,
                past_covariates=[self.covariates, self.covariates],
                allow_model_training=True,
            )

        with pytest.raises(ValueError):
            # number of 'future_covariates' must be the same as the number of Timeseries in 'series'
            anomaly_model.fit(
                series=[self.train, self.train],
                future_covariates=self.covariates,
                allow_model_training=True,
            )

        with pytest.raises(ValueError):
            # number of 'future_covariates' must be the same as the number of Timeseries in 'series'
            anomaly_model.fit(
                series=self.train,
                future_covariates=[self.covariates, self.covariates],
                allow_model_training=True,
            )

    def test_pretrain_forecasting_model(self):
        fitted_model = RegressionModel(lags=10).fit(self.train)
        # Fittable scorer must be fitted before calling .score(), even if forecasting model is fitted
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=fitted_model, scorer=KMeansScorer()).score(
                series=self.test
            )
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=fitted_model, scorer=[Norm(), KMeansScorer()]
            ).score(series=self.test)

        # forecasting model that do not accept past/future covariates
        # with pytest.raises(ValueError):
        #    ForecastingAnomalyModel(model=ExponentialSmoothing(),
        #       scorer=NormScorer()).fit(
        #           series=self.train, past_covariates=self.covariates, allow_model_training=True
        #       )
        # with pytest.raises(ValueError):
        #    ForecastingAnomalyModel(model=ExponentialSmoothing(),
        #       scorer=NormScorer()).fit(
        #           series=self.train, future_covariates=self.covariates, allow_model_training=True
        #       )

        # check window size
        # max window size is len(series.drop_before(series.get_timestamp_at_point(start))) + 1
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=RegressionModel(lags=10),
                scorer=KMeansScorer(window=50, window_agg=False),
            ).fit(series=self.train, start=0.9)

        # forecasting model that cannot be trained on a list of series
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=NaiveSeasonal(), scorer=Norm()).fit(
                series=[self.train, self.train], allow_model_training=True
            )

    @pytest.mark.parametrize("anomaly_model_config", forecasting_am)
    def test_ScoreForecastingAnomalyModelInput(self, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(**am_kwargs)
        anomaly_model.fit(self.train, allow_model_training=True)

        # number of 'past_covariates' must be the same as the number of Timeseries in 'series'
        with pytest.raises(ValueError):
            anomaly_model.score(
                series=[self.train, self.train], past_covariates=self.covariates
            )

        # number of 'past_covariates' must be the same as the number of Timeseries in 'series'
        with pytest.raises(ValueError):
            anomaly_model.score(
                series=self.train,
                past_covariates=[self.covariates, self.covariates],
            )

        # number of 'future_covariates' must be the same as the number of Timeseries in 'series'
        with pytest.raises(ValueError):
            anomaly_model.score(
                series=[self.train, self.train], future_covariates=self.covariates
            )

        # number of 'future_covariates' must be the same as the number of Timeseries in 'series'
        with pytest.raises(ValueError):
            anomaly_model.score(
                series=self.train,
                future_covariates=[self.covariates, self.covariates],
            )

    def test_window_size(self):
        # max window size is len(series.drop_before(series.get_timestamp_at_point(start))) + 1 for score()
        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=10),
            scorer=KMeansScorer(window=30, window_agg=False),
        )
        anomaly_model.fit(self.train, allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.score(series=self.train, start=0.9)

    @pytest.mark.parametrize("anomaly_model_config", filtering_am)
    def test_ScoreFilteringAnomalyModelInput(self, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(**am_kwargs)

        if anomaly_model.scorers_are_trainable:
            anomaly_model.fit(self.train)

    @pytest.mark.parametrize(
        "anomaly_model,fit_kwargs",
        [
            (
                ForecastingAnomalyModel(model=RegressionModel(lags=10), scorer=Norm()),
                {"series": train, "allow_model_training": True},
            ),
            (
                FilteringAnomalyModel(
                    model=MovingAverageFilter(window=20), scorer=Norm()
                ),
                False,
            ),
            (
                ForecastingAnomalyModel(
                    model=RegressionModel(lags=10),
                    scorer=[Norm(), WassersteinScorer(window_agg=False)],
                ),
                {"series": train, "allow_model_training": True},
            ),
            (
                FilteringAnomalyModel(
                    model=MovingAverageFilter(window=20),
                    scorer=[Norm(), WassersteinScorer(window_agg=False)],
                ),
                {"series": train},
            ),
        ],
    )
    def test_eval_metric(self, anomaly_model, fit_kwargs):
        if fit_kwargs:
            anomaly_model.fit(**fit_kwargs)

        # if the anomaly_model have scorers that have the parameter is_univariate set to True,
        # 'anomalies' must have widths of 1
        if anomaly_model.scorers_are_univariate:
            with pytest.raises(ValueError):
                anomaly_model.eval_metric(
                    anomalies=self.mts_anomalies, series=self.test
                )
            with pytest.raises(ValueError):
                anomaly_model.eval_metric(
                    anomalies=self.mts_anomalies, series=self.mts_test
                )
            with pytest.raises(ValueError):
                anomaly_model.eval_metric(
                    anomalies=[self.anomalies, self.mts_anomalies],
                    series=[self.test, self.mts_test],
                )

        # 'metric' must be str and "AUC_ROC" or "AUC_PR"
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies, series=self.test, metric=1
            )
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies,
                series=self.test,
                metric="auc_roc",
            )
        with pytest.raises(TypeError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies,
                series=self.test,
                metric=["AUC_ROC"],
            )

        # 'anomalies' must be binary
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(anomalies=self.test, series=self.test)

        # 'anomalies' must contain anomalies (at least one)
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(anomalies=self.only_0_anomalies, series=self.test)

        # 'anomalies' cannot contain only anomalies
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(anomalies=self.only_1_anomalies, series=self.test)

        # 'anomalies' must match the number of series
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies, series=[self.test, self.test]
            )
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=[self.anomalies, self.anomalies],
                series=self.test,
            )

        # 'anomalies' must have non empty intersection with 'series'
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies[:20], series=self.test[30:]
            )
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=[self.anomalies, self.anomalies[:20]],
                series=[self.test, self.test[40:]],
            )

        # Check input type
        # 'anomalies' and 'series' must be of same length
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies], [self.test, self.test])
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(self.anomalies, [self.test, self.test])
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies, self.anomalies], [self.test])
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies, self.anomalies], self.test)

        # 'anomalies' and 'series' must be of type Timeseries
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies], [2, 3, 4])
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies], "str")
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([2, 3, 4], self.test)
        with pytest.raises(ValueError):
            anomaly_model.eval_metric("str", self.test)
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                [self.anomalies, self.anomalies], [self.test, [3, 2, 1]]
            )
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                [self.anomalies, [3, 2, 1]], [self.test, self.test]
            )

        # Check return types
        # Check if return type is float when input is a series
        assert isinstance(
            anomaly_model.eval_metric(self.anomalies, self.test),
            Dict,
        )

        # Check if return type is Sequence when input is a Sequence of series
        assert isinstance(
            anomaly_model.eval_metric(self.anomalies, [self.test]),
            Sequence,
        )

        assert isinstance(
            anomaly_model.eval_metric(
                [self.anomalies, self.anomalies], [self.test, self.test]
            ),
            Sequence,
        )

    def test_ForecastingAnomalyModelInput(self):
        # model input
        # model input must be of type ForecastingModel
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model="str", scorer=Norm())
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=1, scorer=Norm())
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=MovingAverageFilter(window=10), scorer=Norm())
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=[RegressionModel(lags=10), RegressionModel(lags=5)],
                scorer=Norm(),
            )

        # scorer input
        # scorer input must be of type AnomalyScorer
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=RegressionModel(lags=10), scorer=1)
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=RegressionModel(lags=10), scorer="str")
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=RegressionModel(lags=10), scorer=RegressionModel(lags=10)
            )
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=RegressionModel(lags=10), scorer=[Norm(), "str"]
            )

    def test_FilteringAnomalyModelInput(self):
        # model input
        # model input must be of type FilteringModel
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model="str", scorer=Norm())
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model=1, scorer=Norm())
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model=RegressionModel(lags=10), scorer=Norm())
        with pytest.raises(ValueError):
            FilteringAnomalyModel(
                model=[MovingAverageFilter(window=10), MovingAverageFilter(window=10)],
                scorer=Norm(),
            )

        # scorer input
        # scorer input must be of type AnomalyScorer
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model=MovingAverageFilter(window=10), scorer=1)
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model=MovingAverageFilter(window=10), scorer="str")
        with pytest.raises(ValueError):
            FilteringAnomalyModel(
                model=MovingAverageFilter(window=10),
                scorer=MovingAverageFilter(window=10),
            )
        with pytest.raises(ValueError):
            FilteringAnomalyModel(
                model=MovingAverageFilter(window=10), scorer=[Norm(), "str"]
            )

    def test_univariate_ForecastingAnomalyModel(self):
        np.random.seed(40)

        np_train_slope = np.array(range(0, 100, 1))
        np_test_slope = np.array(range(0, 100, 1))

        np_test_slope[30:32] = 29
        np_test_slope[50:65] = np_test_slope[50:65] + 1
        np_test_slope[75:80] = np_test_slope[75:80] * 0.98

        train_series_slope = TimeSeries.from_values(np_train_slope)
        test_series_slope = TimeSeries.from_values(np_test_slope)

        np_anomalies = np.zeros(100)
        np_anomalies[30:32] = 1
        np_anomalies[50:55] = 1
        np_anomalies[70:80] = 1
        ts_anomalies = TimeSeries.from_times_and_values(
            test_series_slope.time_index, np_anomalies, columns=["is_anomaly"]
        )

        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=5),
            scorer=[
                Norm(),
                Difference(),
                WassersteinScorer(window_agg=False),
                KMeansScorer(k=5),
                KMeansScorer(window=10, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=10, window_agg=False),
                WassersteinScorer(window=15, window_agg=False),
            ],
        )

        anomaly_model.fit(train_series_slope, allow_model_training=True, start=0.1)
        score, pred_series = anomaly_model.score(
            test_series_slope, return_model_prediction=True, start=0.1
        )

        # check that NormScorer is the abs difference of pred_series and test_series_slope
        assert (
            pred_series - test_series_slope.slice_intersect(pred_series)
        ).__abs__() == Norm().score_from_prediction(test_series_slope, pred_series)

        # check that Difference is the difference of pred_series and test_series_slope
        assert test_series_slope.slice_intersect(
            pred_series
        ) - pred_series == Difference().score_from_prediction(
            test_series_slope, pred_series
        )

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, test_series_slope, metric="AUC_ROC", start=0.1
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, test_series_slope, metric="AUC_PR", start=0.1
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            0.773449920508744,
            0.40659777424483307,
            0.9153708133971291,
            0.7702702702702702,
            0.9135765550239234,
            0.7603338632750397,
            0.9153708133971292,
            0.9006591337099811,
        ]

        true_auc_pr = [
            0.4818991248542174,
            0.20023033665128342,
            0.9144135170539835,
            0.47953161438253644,
            0.9127969832903458,
            0.47039678636225957,
            0.9147124232933175,
            0.9604714100445533,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_univariate_FilteringAnomalyModel(self):
        np.random.seed(40)

        np_series_train = np.array(range(0, 100, 1)) + np.random.normal(
            loc=0, scale=1, size=100
        )
        np_series_test = np.array(range(0, 100, 1)) + np.random.normal(
            loc=0, scale=1, size=100
        )

        np_series_test[30:35] = np_series_test[30:35] + np.random.normal(
            loc=0, scale=10, size=5
        )
        np_series_test[50:60] = np_series_test[50:60] + np.random.normal(
            loc=0, scale=4, size=10
        )
        np_series_test[75:80] = np_series_test[75:80] + np.random.normal(
            loc=0, scale=3, size=5
        )

        train_series_noise = TimeSeries.from_values(np_series_train)
        test_series_noise = TimeSeries.from_values(np_series_test)

        np_anomalies = np.zeros(100)
        np_anomalies[30:35] = 1
        np_anomalies[50:60] = 1
        np_anomalies[75:80] = 1
        ts_anomalies = TimeSeries.from_times_and_values(
            test_series_noise.time_index, np_anomalies, columns=["is_anomaly"]
        )

        anomaly_model = FilteringAnomalyModel(
            model=MovingAverageFilter(window=5),
            scorer=[
                Norm(),
                Difference(),
                WassersteinScorer(window_agg=False),
                KMeansScorer(),
                KMeansScorer(window=10, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=10, window_agg=False),
                WassersteinScorer(window=15, window_agg=False),
            ],
        )
        anomaly_model.fit(train_series_noise)
        score, pred_series = anomaly_model.score(
            test_series_noise, return_model_prediction=True
        )

        # check that Difference is the difference of pred_series and test_series_noise
        assert test_series_noise.slice_intersect(
            pred_series
        ) - pred_series == Difference().score_from_prediction(
            test_series_noise, pred_series
        )

        # check that NormScorer is the abs difference of pred_series and test_series_noise
        assert (
            test_series_noise.slice_intersect(pred_series) - pred_series
        ).__abs__() == Norm().score_from_prediction(test_series_noise, pred_series)

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, test_series_noise, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, test_series_noise, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            0.875625,
            0.5850000000000001,
            0.952127659574468,
            0.814375,
            0.9598646034816247,
            0.88125,
            0.9666344294003868,
            0.9731182795698925,
        ]

        true_auc_pr = [
            0.7691407907338141,
            0.5566414178265074,
            0.9720504927710986,
            0.741298584352156,
            0.9744855592642071,
            0.7808056518442923,
            0.9800621192517156,
            0.9911842778990486,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_univariate_covariate_ForecastingAnomalyModel(self):
        np.random.seed(40)

        day_week = [0, 1, 2, 3, 4, 5, 6]
        np_day_week = np.array(day_week * 10)

        np_train_series = 0.5 * np_day_week
        np_test_series = 0.5 * np_day_week

        np_test_series[30:35] = np_test_series[30:35] + np.random.normal(
            loc=0, scale=2, size=5
        )
        np_test_series[50:60] = np_test_series[50:60] + np.random.normal(
            loc=0, scale=1, size=10
        )

        covariates = TimeSeries.from_times_and_values(
            pd.date_range(start="1949-01-01", end="1949-03-11"), np_day_week
        )
        series_train = TimeSeries.from_times_and_values(
            pd.date_range(start="1949-01-01", end="1949-03-11"), np_train_series
        )
        series_test = TimeSeries.from_times_and_values(
            pd.date_range(start="1949-01-01", end="1949-03-11"), np_test_series
        )

        np_anomalies = np.zeros(70)
        np_anomalies[30:35] = 1
        np_anomalies[50:60] = 1
        ts_anomalies = TimeSeries.from_times_and_values(
            series_test.time_index, np_anomalies, columns=["is_anomaly"]
        )

        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=2, lags_future_covariates=[0]),
            scorer=[
                Norm(),
                Difference(),
                WassersteinScorer(window_agg=False),
                KMeansScorer(k=4),
                KMeansScorer(k=7, window=10, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=10, window_agg=False),
                WassersteinScorer(window=15, window_agg=False),
            ],
        )

        anomaly_model.fit(
            series_train,
            allow_model_training=True,
            future_covariates=covariates,
            start=0.2,
        )

        score, pred_series = anomaly_model.score(
            series_test,
            return_model_prediction=True,
            future_covariates=covariates,
            start=0.2,
        )

        # check that NormScorer is the abs difference of pred_series and series_test
        assert (
            series_test.slice_intersect(pred_series) - pred_series
        ).__abs__() == Norm().score_from_prediction(series_test, pred_series)

        # check that Difference is the difference of pred_series and series_test
        assert series_test.slice_intersect(
            pred_series
        ) - pred_series == Difference().score_from_prediction(series_test, pred_series)

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, series_test, metric="AUC_ROC", start=0.2
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, series_test, metric="AUC_PR", start=0.2
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [1.0, 0.6, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]

        true_auc_pr = [
            1.0,
            0.6914399076961142,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            0.9999999999999999,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_multivariate_FilteringAnomalyModel(self):
        np.random.seed(40)

        data_1 = np.random.normal(0, 0.1, 100)
        data_2 = np.random.normal(0, 0.1, 100)

        mts_series_train = TimeSeries.from_values(
            np.dstack((data_1, data_2))[0], columns=["component 1", "component 2"]
        )

        data_1[15:20] = data_1[15:20] + np.random.normal(0, 0.9, 5)
        data_1[35:40] = data_1[35:40] + np.random.normal(0, 0.4, 5)

        data_2[50:55] = data_2[50:55] + np.random.normal(0, 0.7, 5)
        data_2[65:70] = data_2[65:70] + np.random.normal(0, 0.4, 5)

        data_1[80:85] = data_1[80:85] + np.random.normal(0, 0.6, 5)
        data_2[80:85] = data_2[80:85] + np.random.normal(0, 0.6, 5)

        data_1[93:98] = data_1[93:98] + np.random.normal(0, 0.6, 5)
        data_2[93:98] = data_2[93:98] + np.random.normal(0, 0.6, 5)
        mts_series_test = TimeSeries.from_values(
            np.dstack((data_1, data_2))[0], columns=["component 1", "component 2"]
        )

        np1_anomalies = np.zeros(len(data_1))
        np1_anomalies[15:20] = 1
        np1_anomalies[35:40] = 1
        np1_anomalies[80:85] = 1
        np1_anomalies[93:98] = 1

        np2_anomalies = np.zeros(len(data_2))
        np2_anomalies[50:55] = 1
        np2_anomalies[67:70] = 1
        np2_anomalies[80:85] = 1
        np2_anomalies[93:98] = 1

        np_anomalies = np.zeros(len(data_2))
        np_anomalies[15:20] = 1
        np_anomalies[35:40] = 1
        np_anomalies[50:55] = 1
        np_anomalies[67:70] = 1
        np_anomalies[80:85] = 1
        np_anomalies[93:98] = 1

        ts_anomalies = TimeSeries.from_times_and_values(
            mts_series_train.time_index,
            np.dstack((np1_anomalies, np2_anomalies))[0],
            columns=["is_anomaly_1", "is_anomaly_2"],
        )

        mts_anomalies = TimeSeries.from_times_and_values(
            mts_series_train.time_index, np_anomalies, columns=["is_anomaly"]
        )

        # first case: scorers that return univariate scores
        anomaly_model = FilteringAnomalyModel(
            model=MovingAverageFilter(window=10),
            scorer=[
                Norm(component_wise=False),
                WassersteinScorer(window_agg=False),
                WassersteinScorer(window=12, window_agg=False),
                KMeansScorer(),
                KMeansScorer(window=5, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=5, window_agg=False),
            ],
        )
        anomaly_model.fit(mts_series_train)

        scores, pred_series = anomaly_model.score(
            mts_series_test, return_model_prediction=True
        )

        # pred_series must be multivariate (same width as input)
        assert pred_series.width == mts_series_test.width

        # scores must be of the same length as the number of scorers
        assert len(scores) == len(anomaly_model.scorers)

        dict_auc_roc = anomaly_model.eval_metric(
            mts_anomalies, mts_series_test, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            mts_anomalies, mts_series_test, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[mts_anomalies] * 7,
            pred_scores=scores,
            window=[1, 10, 12, 1, 5, 1, 5],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[mts_anomalies] * 7,
            pred_scores=scores,
            window=[1, 10, 12, 1, 5, 1, 5],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            0.8695436507936507,
            0.9737678855325913,
            0.9930555555555555,
            0.857638888888889,
            0.9639130434782609,
            0.8690476190476191,
            0.9630434782608696,
        ]

        true_auc_pr = [
            0.814256917602188,
            0.9945160041091712,
            0.9992086070916503,
            0.8054288542539664,
            0.9777504211642852,
            0.8164636240285442,
            0.9763049418985656,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

        # second case: scorers that return scorers that have the same width as the input
        anomaly_model = FilteringAnomalyModel(
            model=MovingAverageFilter(window=10),
            scorer=[
                Norm(component_wise=True),
                Difference(),
                WassersteinScorer(component_wise=True, window_agg=False),
                WassersteinScorer(window=12, component_wise=True, window_agg=False),
                KMeansScorer(component_wise=True),
                KMeansScorer(window=5, component_wise=True, window_agg=False),
                PyODScorer(model=KNN(), component_wise=True),
                PyODScorer(
                    model=KNN(), window=5, component_wise=True, window_agg=False
                ),
            ],
        )
        anomaly_model.fit(mts_series_train)

        scores, pred_series = anomaly_model.score(
            mts_series_test, return_model_prediction=True
        )

        # pred_series must be multivariate (same width as input)
        assert pred_series.width == mts_series_test.width

        # scores must be of the same length as the number of scorers
        assert len(scores) == len(anomaly_model.scorers)

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, mts_series_test, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, mts_series_test, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=scores,
            window=[1, 1, 10, 12, 1, 5, 1, 5],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=scores,
            window=[1, 1, 10, 12, 1, 5, 1, 5],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            [0.859375, 0.9200542005420054],
            [0.49875, 0.513550135501355],
            [0.997093023255814, 0.9536231884057971],
            [0.998960498960499, 0.9739795918367344],
            [0.8143750000000001, 0.8218157181571816],
            [0.9886148007590132, 0.94677734375],
            [0.830625, 0.9369918699186992],
            [0.9909867172675522, 0.94580078125],
        ]

        true_auc_pr = [
            [0.7213314465376244, 0.8191331553279771],
            [0.4172305056124696, 0.49249755343619195],
            [0.9975245098039216, 0.9741870252257915],
            [0.9992877492877493, 0.9865792868871687],
            [0.7095552075210219, 0.7591858780309868],
            [0.9827224901431558, 0.9402925739221939],
            [0.7095275592303261, 0.8313668186652059],
            [0.9858096294704315, 0.9391783485106905],
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_multivariate_ForecastingAnomalyModel(self):
        np.random.seed(40)

        data_sin = np.array([np.sin(x) for x in np.arange(0, 20 * np.pi, 0.2)])
        data_cos = np.array([np.cos(x) for x in np.arange(0, 20 * np.pi, 0.2)])

        mts_series_train = TimeSeries.from_values(
            np.dstack((data_sin, data_cos))[0], columns=["component 1", "component 2"]
        )

        data_sin[10:20] = 0
        data_cos[60:80] = 0

        data_sin[100:110] = 1
        data_cos[150:155] = 1

        data_sin[200:240] = 0.9 * data_cos[200:240]
        data_cos[200:240] = 0.9 * data_sin[200:240]

        data_sin[275:295] = data_sin[275:295] + np.random.normal(0, 0.1, 20)
        data_cos[275:295] = data_cos[275:295] + np.random.normal(0, 0.1, 20)

        mts_series_test = TimeSeries.from_values(
            np.dstack((data_sin, data_cos))[0], columns=["component 1", "component 2"]
        )

        np1_anomalies = np.zeros(len(data_sin))
        np1_anomalies[10:20] = 1
        np1_anomalies[100:110] = 1
        np1_anomalies[200:240] = 1
        np1_anomalies[275:295] = 1

        np2_anomalies = np.zeros(len(data_cos))
        np2_anomalies[60:80] = 1
        np2_anomalies[150:155] = 1
        np2_anomalies[200:240] = 1
        np2_anomalies[275:295] = 1

        np_anomalies = np.zeros(len(data_cos))
        np_anomalies[10:20] = 1
        np_anomalies[60:80] = 1
        np_anomalies[100:110] = 1
        np_anomalies[150:155] = 1
        np_anomalies[200:240] = 1
        np_anomalies[275:295] = 1

        ts_anomalies = TimeSeries.from_times_and_values(
            mts_series_train.time_index,
            np.dstack((np1_anomalies, np2_anomalies))[0],
            columns=["is_anomaly_1", "is_anomaly_2"],
        )

        mts_anomalies = TimeSeries.from_times_and_values(
            mts_series_train.time_index, np_anomalies, columns=["is_anomaly"]
        )

        # first case: scorers that return univariate scores
        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=10),
            scorer=[
                Norm(component_wise=False),
                WassersteinScorer(window_agg=False),
                WassersteinScorer(window=20, window_agg=False),
                KMeansScorer(),
                KMeansScorer(window=20, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=10, window_agg=False),
            ],
        )
        anomaly_model.fit(mts_series_train, allow_model_training=True, start=0.1)

        scores, pred_series = anomaly_model.score(
            mts_series_test, return_model_prediction=True, start=0.1
        )

        # pred_series must be multivariate (same width as input)
        assert pred_series.width == mts_series_test.width

        # scores must be of the same length as the number of scorers
        assert len(scores) == len(anomaly_model.scorers)

        dict_auc_roc = anomaly_model.eval_metric(
            mts_anomalies, mts_series_test, start=0.1, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            mts_anomalies, mts_series_test, start=0.1, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[mts_anomalies] * 7,
            pred_scores=scores,
            window=[1, 10, 20, 1, 20, 1, 10],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[mts_anomalies] * 7,
            pred_scores=scores,
            window=[1, 10, 20, 1, 20, 1, 10],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            0.9252575884154831,
            0.9130158730158731,
            0.9291228070175439,
            0.9252575884154832,
            0.9211929824561403,
            0.9252575884154831,
            0.915873015873016,
        ]

        true_auc_pr = [
            0.8389462532437767,
            0.9151621069238896,
            0.9685249535885079,
            0.8389462532437765,
            0.9662153835545242,
            0.8389462532437764,
            0.9212725256428517,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

        # second case: scorers that return scorers that have the same width as the input
        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=10),
            scorer=[
                Norm(component_wise=True),
                Difference(),
                WassersteinScorer(component_wise=True, window_agg=False),
                WassersteinScorer(window=20, component_wise=True, window_agg=False),
                KMeansScorer(component_wise=True),
                KMeansScorer(window=20, component_wise=True, window_agg=False),
                PyODScorer(model=KNN(), component_wise=True),
                PyODScorer(
                    model=KNN(), window=10, component_wise=True, window_agg=False
                ),
            ],
        )
        anomaly_model.fit(mts_series_train, allow_model_training=True, start=0.1)

        scores, pred_series = anomaly_model.score(
            mts_series_test, return_model_prediction=True, start=0.1
        )

        # pred_series must be multivariate (same width as input)
        assert pred_series.width == mts_series_test.width

        # scores must be of the same length as the number of scorers
        assert len(scores) == len(anomaly_model.scorers)

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, mts_series_test, start=0.1, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, mts_series_test, start=0.1, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=scores,
            window=[1, 1, 10, 20, 1, 20, 1, 10],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=scores,
            window=[1, 1, 10, 20, 1, 20, 1, 10],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            [0.8803738317757009, 0.912267218445167],
            [0.48898531375166887, 0.5758202778598878],
            [0.8375999073323295, 0.9162283996994741],
            [0.7798128494807715, 0.8739249880554228],
            [0.8803738317757008, 0.912267218445167],
            [0.7787287458632889, 0.8633540372670807],
            [0.8803738317757009, 0.9122672184451671],
            [0.8348777945094406, 0.9137061285821616],
        ]

        true_auc_pr = [
            [0.7123114333965317, 0.7579757115620807],
            [0.4447973021706103, 0.596776950584551],
            [0.744325434474558, 0.8984960888744328],
            [0.7653561450296187, 0.9233662817550338],
            [0.7123114333965317, 0.7579757115620807],
            [0.7852553779986415, 0.9185701347601994],
            [0.7123114333965319, 0.7579757115620807],
            [0.757208451057927, 0.8967178983419622],
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_visualization(self):
        # test function show_anomalies() and show_anomalies_from_scores()
        forecasting_anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=10), scorer=Norm()
        )
        forecasting_anomaly_model.fit(self.train, allow_model_training=True)

        filtering_anomaly_model = FilteringAnomalyModel(
            model=MovingAverageFilter(window=10), scorer=Norm()
        )

        self.show_anomalies_function(
            visualization_function=forecasting_anomaly_model.show_anomalies
        )
        self.show_anomalies_function(
            visualization_function=filtering_anomaly_model.show_anomalies
        )
        self.show_anomalies_function(visualization_function=show_anomalies_from_scores)

    def show_anomalies_function(self, visualization_function):
        # must input only one series
        with pytest.raises(ValueError) as err:
            visualization_function(series=[self.train, self.train])
        assert (
            str(err.value)
            == "`series` must be single `TimeSeries` or a sequence of `TimeSeries` of length `1`."
        )
        # input must be a series
        with pytest.raises(ValueError):
            visualization_function(series=[1, 2, 4])

        if visualization_function != show_anomalies_from_scores:
            # metric must be "AUC_ROC" or "AUC_PR"
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.anomalies,
                    metric="str",
                )
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.anomalies,
                    metric="auc_roc",
                )
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train, anomalies=self.anomalies, metric=1
                )

            # anomalies must be not none if metric is given
            with pytest.raises(ValueError):
                visualization_function(series=self.train, metric="AUC_ROC")

            # anomalies must be binary
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.test,
                    metric="AUC_ROC",
                )

            # anomalies must contain at least 1 anomaly if metric is given
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.only_0_anomalies,
                    metric="AUC_ROC",
                )

            # anomalies must contain at least 1 non-anomoulous timestamp
            # if metric is given
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.only_1_anomalies,
                    metric="AUC_ROC",
                )
        else:
            # window must be a positive int
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train, pred_scores=self.test, window=-1
                )
            # window must smaller than the score series
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train, pred_scores=self.test, window=200
                )
            # must have the same nbr of windows than scores
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train, pred_scores=self.test, window=[1, 2]
                )
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train,
                    pred_scores=[self.test, self.test],
                    window=[1, 2, 1],
                )
            # nbr of names_of_scorers must match the nbr of scores
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train,
                    pred_scores=self.test,
                    names_of_scorers=["scorer1", "scorer2"],
                )
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train,
                    pred_scores=[self.test, self.test],
                    names_of_scorers=["scorer1", "scorer2", "scorer3"],
                )



"""
Threshold Detector
------------------

Detector that detects anomaly based on user-given threshold.
This detector compares time series values with user-given thresholds, and
identifies time points as anomalous when values are beyond the thresholds.
"""

from typing import Sequence, Union

import numpy as np

from darts.ad.detectors.detectors import Detector, _BoundedDetectorMixin
from darts.logging import get_logger, raise_log
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class ThresholdDetector(Detector, _BoundedDetectorMixin):
    def __init__(
        self,
        low_threshold: Union[int, float, Sequence[float], None] = None,
        high_threshold: Union[int, float, Sequence[float], None] = None,
    ) -> None:
        """Threshold Detector

        Flags values that are either below or above the `low_threshold` and `high_threshold`,
        respectively.

        If a single value is provided for `low_threshold` or `high_threshold`, this same
        value will be used across all components of the series.

        If sequences of values are given for the parameters `low_threshold` and/or `high_threshold`,
        they must be of the same length, matching the dimensionality of the series passed
        to `detect()`, or have a length of 1. In the latter case, this single value will be used
        across all components of the series.

        If either `low_threshold` or `high_threshold` is None, the corresponding bound will not be used.
        However, at least one of the two must be set.

        Parameters
        ----------
        low_threshold
            (Sequence of) lower bounds. If a sequence, must match the dimensionality of the series this
            detector is applied to.
        high_threshold
            (Sequence of) upper bounds. If a sequence, must match the dimensionality of the series this
            detector is applied to.
        """
        super().__init__()
        low_threshold, high_threshold = self._prepare_boundaries(
            lower_bound=low_threshold,
            upper_bound=high_threshold,
            lower_bound_name="low_threshold",
            upper_bound_name="high_threshold",
        )
        self._low_threshold = low_threshold
        self._high_threshold = high_threshold

    def _detect_core(self, series: TimeSeries, name: str = "series") -> TimeSeries:
        if len(self.low_threshold) > 1 and len(self.low_threshold) != series.width:
            raise_log(
                ValueError(
                    f"The number of components for each series in `{name}` must be "
                    f"equal to the number of threshold values. Found number of "
                    f"components equal to {series.width} and expected {len(self.low_threshold)}."
                ),
                logger=logger,
            )

        # if length is 1, tile it to series width:
        low_threshold = self._expand_threshold(series[0], self.low_threshold)
        high_threshold = self._expand_threshold(series[0], self.high_threshold)

        # (time, components)
        np_series = series.values(copy=False)

        def _detect_fn(x, lo, hi):
            # x of shape (time,) for 1 component
            return (x < (np.NINF if lo is None else lo)) | (
                x > (np.Inf if hi is None else hi)
            )

        detected = np.zeros_like(np_series, dtype=int)

        for component_idx in range(series.width):
            detected[:, component_idx] = _detect_fn(
                np_series[:, component_idx],
                low_threshold[component_idx],
                high_threshold[component_idx],
            )
        return series.with_values(np.expand_dims(detected, -1).astype(series.dtype))

    @property
    def low_threshold(self):
        return self._low_threshold

    @property
    def high_threshold(self):
        return self._high_threshold
